;; Generated package description from wpuzzle.el  -*- no-byte-compile: t -*-
(define-package "wpuzzle" "1.1.0.20201128.144116" "find as many word in a given time" 'nil :url "https://elpa.gnu.org/packages/wpuzzle.html" :authors '(("Ivan Kanis" . "ivan@kanis.fr")) :maintainer '("Ivan Kanis" . "ivan@kanis.fr"))
